function [team, cc_score, time_forte] = forte(W, M_lbs, lbs, M_ubs, ubs, g, S, dist, d_0, options)
% Finds a team of experts that sasfies the given constraints maximimizing the collaborative compatibility 
% 
% Usage:
% [team, cc_score, time_forte] = forte(W, M_lbs, lbs, M_ubs, ubs, g, S, dist, d_0, options)
%
% Input:
%   W:          Symmetric (SPARSE) weight matrix with the (i,j)-th entry specifiying the compatability between experts i and j
%   M_lbs:      Skill matrix for the lower bound constraints and has dimension, n times m, where n is the total number of experts and m is the total number of skills.
%               Each column j of M_lbs corresponds to a skill and (i,j)-th entry specifies the skill level of expert i in skill j
%   lbs:        Lower bounds for each skill 
%   M_ubs:      Skill matrix for the upper bound constraints and has dimension, n times m, where n is the total number of experts and m is the total number of skills.
%               Each column j of M_lbs corresponds to a skill and (i,j)-th entry specifies the skill level of expert i in skill j
%   ubs:        Upper bounds for each skill
%   g:          Vertex weights used in the definition of (generalzied version of) density
%
% Optional Arguments
%   S:              Set of experts that should be included in the team. Default value: empty set
%   dist:           Distance matrix for the distance-based constraint. Default value: sparse(n, n)
%   d_0:            Upper bound on the distance between any two experts. Default value: zero
%   options:        Use this structure to provide additional options to the method, e.g, providing initialization. 
%       MAXITER:        Maximum of number of iterations for the inner problem. Default value: 1000
%       perturbation:   Set this flag for improved results at the expense of computational time. Default value: true
%       tol_score:      Numerical tolerance for the compatibility score (used to exit the problem). Default value: 1e-4
%       debug:          If enabled prints debugging messages. Default value: false
%       initialization: Warm-start for the method; e.g., solution of the greedy method. When no initialization is specified the method is started with the unconstrained solution 
%                       (i.e., global optimal solution of the densest subgraph problem).
% Output:
%   team:       0/1 vector of size n. One indicates that the presence of the expert in the final team.
%   cc_score:   Collaborative compatibility score of the team found
%   time_forte: Total time taken by the method.
%
% (C)2012-13 Syama Sundar Rangapuram, Thomas Buehler and Matthias Hein 
% Machine Learning Group, Saarland University, Germany
% http://www.ml.uni-saarland.de


    assert(sum(diag(W))==0, 'No self loops: diagonal of the weight matrix W must be zero.');
    assert(issparse(W), 'The weight matrix W must be sparse');
    assert(issymmetric(W), 'The weight matrix W must be symmetric');
    
    if ~exist('S', 'var')
        S = [];
    end
    
    if ~exist('d_0', 'var')
        d_0 = 0;
    end
    
    if ~exist('options', 'var')
        options.MAXITER = 1000;
        options.perturbation = true;
        options.tol_score = 1e-4;
        options.debug = false;
        options.initialization = [];
    else
        if ~isfield(options, 'MAXITER')
            options.MAXITER = 1000;
        end
        if ~isfield(options, 'perturbation')
            options.perturbation = true;
        end
        if ~isfield(options, 'tol_score')
            options.tol_score = true;
        end
        if ~isfield(options, 'debug')
            options.debug = false;
        end
        if ~isfield(options, 'initialization')
            options.initialization = [];
        end
    end
    
    
    disp('************* Forte: Solving an instance of the team formation problem **************');
    
    if ~isempty(S) && d_0 > 0 % We have seed and distance constraints, so we solve an equivalent problem on the reduced graph.
        % Experts that are at a distance of larger than d_0 from any expert in S can be eliminated without changing the problem
        ix_reduced = find(sum(dist(:, S)<=d_0, 2)==length(S)); % the distance should be smaller or equal to each of the expert in S.
        S_reduced = find(ismember(ix_reduced, S));      % S_reduced is the index of S in these indices.
        W_reduced = W(ix_reduced, ix_reduced);
        dist_reduced = dist(ix_reduced, ix_reduced);
        g_reduced = g(ix_reduced);
        M_lbs_reduced = M_lbs(ix_reduced, :);
        M_ubs_reduced = M_ubs(ix_reduced, :); 
        if ~isempty(options.initialization)
            options.initialization = options.initialization(ix_reduced);
        end
        clear dist;

        %dist_reduced(dist_reduced<=d_0) = 0;
        [team_forte_reduced, cc_score, time_forte] = forte_standard_form(W_reduced, M_lbs_reduced, lbs, M_ubs_reduced, ubs, g_reduced, S_reduced, dist_reduced, d_0, options.MAXITER, options.perturbation, options.tol_score, options.initialization, options.debug);     
        n = size(W,1);
        team = sparse(n,1);
        team(ix_reduced(team_forte_reduced>0)) = 1; 
    else
        [team, cc_score, time_forte] = forte_standard_form(W, M_lbs, lbs, M_ubs, ubs, g, S, dist, d_0, options.MAXITER, options.perturbation, options.tol_score, options.initialization, options.debug);
    end

end

function [team, cc_score, time_forte] = forte_standard_form(W, M_lbs, lbs, M_ubs, ubs, g, S, dist, d_0, MAXITER, perturbation, tol_score, starts, debug)

    if min(min(M_lbs))>=0
        max_M_lbs = max(M_lbs);
        max_M_lbs(max_M_lbs==0) = 1; % no division by zero!
        M_lbs = bsxfun(@rdivide, M_lbs, max_M_lbs);
        lbs = lbs./max_M_lbs;
    end

    if min(min(M_ubs))>=0
        max_M_ubs = max(M_ubs);
        max_M_ubs(max_M_ubs==0) = 1; % no division by zero!
        M_ubs = bsxfun(@rdivide, M_ubs, max_M_ubs);
        ubs = ubs./max_M_ubs;
    end

    % shift the distance constraint
    dist = sparse(max(0, dist - d_0));
    d_0 = 0;
    % Rescale the distance constraint so that all distances are between [0,1]
    max_dist = max(max(dist));
    if max_dist>0, dist = dist/max_dist; end

    norm_penalties = 0;
    n = size(W,1);
    %m = size(M_lbs,2);
    m = length(lbs) + length(ubs)+ 1;% (sum(sum(dist))>0);
    gamma_counter = 0.1;
    gammas = 0;
    %pert_probs = [0.01:0.01:0.1 0.2:0.1:1];
    pert_probs = [0.01 0.02];

    satisfied = false;

    initialized = false;
    nRuns = 1;
    if isempty(starts)
        starts = randn(n, nRuns);
    else
        initialized = true;
        if debug
            display('Forte with initialization');
        end
    end

    objectives = zeros(nRuns, length(gammas));
    viols = zeros(nRuns, length(gammas));
    solutions = zeros(n, nRuns, length(gammas));
    ip_opt_solutions = zeros(nRuns, length(gammas));
    prev_viols = m;

    gammas_final = zeros(m, 1);
    final_ixs = [];
    % This is for one instance of the team formation problem
    time_forte_start = tic;
    while ~satisfied % while not all constraints are satisfied... gamma loop
        %it is advisable to have gamma iteraion as the outer loop because it allows to stop early
        % if we find very good solution for smaller gamma without
        % exploring the gamma path for other starts

        gamma = gammas(end);
        it = 1; sol_found = false; proj_counter = 1;
        infeas_gaps_additional = []; additional_viols = [];
        while ~sol_found;

            if gamma == 0
                if ~isempty(S)
                    [team,cc_score,objective,ip_opt]= forte_soft_cnstrs(W,zeros(1, length(ubs)),zeros(1, length(lbs)),g,zeros(n,length(ubs)), zeros(n, length(lbs)), S, sparse(n,n), 0, starts(:, it),MAXITER,gamma, sparse(m,1), tol_score, debug);
                else
                    [team, cc_score, objective]= densest_subgraph(W, g, tol_score, debug);
                    ip_opt = true;
                end
                last_cc_score = cc_score;
            else
                cur_f = prev_soln;  % cur_f is used to find gamma scales for the penalty terms.
                infeas_gaps = [max(0, lbs - (cur_f>0)'*M_lbs) max(0, (cur_f>0)'*M_ubs - ubs) sum(sum( max(0, dist(cur_f>0, cur_f>0) - d_0) ))];
                infeas_gaps(additional_viols) = infeas_gaps_additional;
                infeas_gaps_px = infeas_gaps'/sum(infeas_gaps);

                %norm_penalties = (gammas(end-1) * (norm_penalties>0) + gamma_counter * norm_penalties) / gamma;
                norm_penalties = ( gammas_final + gamma_counter * infeas_gaps_px) / gamma;

                uniq_values = unique(starts(:, it));
                if length(uniq_values) == 2
                    C = find(starts(:, it));
                    deg_C = sum(W(C,C),2);
                    start = starts(:, it);

                    start(C) = deg_C - (1/objectives(it, end))*g(C);
                    start(C) = start(C) - min(start(C)) + 1e-6;
                    assert(sum(start(C)<0)==0);

                    start = start/sum(start);

                    for lv=1:length(lbs)
                        if infeas_gaps(lv)
                            lv_mem = find(M_lbs(:, lv));
                            deg_C = sum(W(lv_mem, cur_f>0), 2); % degree of each member to the current team.

                            if sum(deg_C) ~=0, deg_C_px = deg_C/sum(deg_C); end
                            start(lv_mem) = 0.1*deg_C_px;
                        end
                    end

                    starts(:, it) = start;
                    %display(num2str(C'));
                    %display(num2str(start(C)'));

                    start1 = start;

                    start(C) = sum(deg_C)./g(C);
                    for skill=1:size(M_lbs,2)
                        Cskill = C(M_lbs(C, skill)>0);
                        start(Cskill) = start(Cskill)/sum(start(Cskill));
                    end
                end
                [team,cc_score,objective,ip_opt]= forte_soft_cnstrs(W,ubs,lbs,g,M_ubs, M_lbs,S,dist, d_0, start1 ,MAXITER,gamma, norm_penalties, tol_score, debug);
            end

            if perturbation && cc_score < (last_cc_score - 1e-8) % no perturbation if the team is not changed!

                disp(['Before perturbation: Found a team of size ', num2str(sum(team)), ' with score: ', num2str(generalized_density(W, g, team))]);        
                display(['Before perturbation: Score (including the penalty term): ', num2str(1/objective)]);
                
                for p=1:length(pert_probs) % try different perturbation probabilities

                    pert_counter = 0;
                    while pert_counter < 1 % for each prob, try until a perturbation does not improve the result

                        display(['Perturbing the solution with noise: ', num2str(pert_probs(p))]);
                        pert_start = team + pert_probs(p)* range(team) * rand(n,1);
                        [team_pert,cc_score_pert,objective_pert,ip_opt_pert]= forte_soft_cnstrs(W,ubs,lbs,g,M_ubs, M_lbs,S,dist, d_0, pert_start ,MAXITER,gamma, norm_penalties, tol_score, debug);
                        
                        if objective_pert < objective && cc_score_pert >= cc_score && full(sum(sum(W(team_pert>0, team_pert>0))))*(objective_pert - 1/cc_score_pert) <= full(sum(sum(W(team>0, team>0))))*(objective - 1/cc_score)
                            display(['Before perturbation: Score: ', num2str(cc_score), ' Penalty: ', num2str(objective - 1/cc_score)]);
                            display(['After perturbation: Score: ', num2str(cc_score_pert), ' Penalty: ', num2str(objective_pert - 1/cc_score_pert)]);
                            if sum( team'*M_lbs < lbs )
                                display(['Before perturbation: lower bound constraint is violated: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
                            else
                                display(['Before perturbation: lower bound constraint is satisfied: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
                            end
                            if sum( team'*M_ubs > ubs )
                                display(['Before perturbation: upper bound constraint is violated: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
                            else
                                display(['Before perturbation: upper bound constraint is satisfied: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
                            end
                            if sum(sum( max(0, dist(team>0, team>0) - d_0) )) ~= 0  %max(max(dist(team>0, team>0))) ~= 0
                                display(['Before perturbation: (Shifted/Scaled) Distance constraint is violated: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
                            else
                                display(['Before perturbation: (Shifted/Scaled) Distance constraint is satisfied: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
                            end
                            
                            %if debug
                                disp(['Perturbation improves the score (including the penalty term) from ', num2str(1/objective), ' to ', num2str(1/objective_pert)]);
                            %end
                            team = team_pert;
                            cc_score = cc_score_pert;
                            objective = objective_pert;
                            ip_opt = ip_opt_pert;

                            % reset the pert_counter
                            %pert_counter = 0;
                            
                        else
                            pert_counter = pert_counter + 1;
                            if debug
                                disp(['Perturbation does not improve the result. Perturbed score (including the penalty term) is ', num2str(1/objective_pert), '. Original was ', num2str(1/objective)]);
                            end

                        end
                    end

                end

            end

            % Now save the results for the current gamma... we have
            % nRuns solutions
            solutions(:, it, length(gammas)) = team;
            ip_opt_solutions(it, length(gammas)) = ip_opt;
            objectives(it, length(gammas)) = objective;
            %all_viols = [ max(0, lbs - team'*M_lbs) max(0, team'*M_ubs - ubs)  sum(sum(dist(team>0, team>0)))];
            all_viols = [ max(0, lbs - team'*M_lbs) max(0, team'*M_ubs - ubs)  sum(sum( max(0, dist(team>0, team>0) - d_0) ))];
            %viols(it, length(gammas)) = sum( team'*M_lbs < lbs ) + sum(team'*M_ubs > ubs) + max(max(dist(team>0, team>0)));
            viols(it, length(gammas)) = sum( team'*M_lbs < lbs ) + sum(team'*M_ubs > ubs) + sum(sum( max(0, dist(team>0, team>0) - d_0) ));

            new_viols_lbs = find(team'*M_lbs < lbs ); new_viols_ubs = find(team'*M_ubs > ubs);
            new_viols_dist = sum(sum( max(0, dist(team>0, team>0) - d_0) ));%new_viols_dist = sum(sum(dist(team>0, team>0)));
            new_viols = [new_viols_lbs length(lbs)+new_viols_ubs];
            if new_viols_dist > 0
                new_viols = [new_viols length(lbs)+length(ubs)+1];
            end
            additional_viols = intersect(final_ixs, new_viols);

            if ~isempty(additional_viols)  %(intersect(final_ixs, new_viols))        % a new constraint is violated!
                %gammas_final( new_viols ) = gamma*proj_counter;
                gammas_final( additional_viols ) = gammas(end-1)*proj_counter;
                proj_counter = proj_counter*2;

                infeas_gaps_additional = all_viols(additional_viols);
                final_ixs = setdiff(final_ixs, new_viols);
                disp(['Found a team of size ', num2str(sum(team)), ' with score: ', num2str(generalized_density(W, g, team))]);        
                if debug
                    if sum( team'*M_lbs < lbs )
                        display(['lower bound constraint is violated: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
                    else
                        display(['lower bound constraint is satisfied: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
                    end
                    if sum( team'*M_ubs > ubs )
                        display(['upper bound constraint is violated: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
                    else
                        display(['upper bound constraint is satisfied: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
                    end
                    if sum(sum( max(0, dist(team>0, team>0) - d_0) )) ~= 0  %max(max(dist(team>0, team>0))) ~= 0
                        display(['(Shifted/Scaled) Distance constraint is violated: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
                    else
                        display(['(Shifted/Scaled) Distance constraint is satisfied: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
                    end
                    display('continue');                    
                end            
                continue;
            end

            proj_counter = 1;
            sol_found = true;
            prev_soln = team;
            new_final_ixs_lbs = setdiff(find(team'*M_lbs >= lbs), final_ixs);
            %gammas_final(team'*M_lbs < lbs) = gamma;
            gammas_final(new_viols_lbs) = gamma;
            gammas_final( new_final_ixs_lbs ) = gamma;

            new_final_ixs_ubs = setdiff( length(lbs)+find(team'*M_ubs <= ubs), final_ixs);
            %gammas_final(length(lbs)+ find(team'*M_ubs > ubs) ) = gamma;
            gammas_final(length(lbs)+ new_viols_ubs ) = gamma;
            gammas_final( new_final_ixs_ubs ) = gamma;

            final_ixs = [final_ixs new_final_ixs_lbs new_final_ixs_ubs];

            if all_viols(end)   % distance constraint is violated
                gammas_final(end) = gamma;
            else
                final_ixs = union(final_ixs, length(all_viols));
            end
            
            disp(['Found a team of size ', num2str(sum(team)), ' with score: ', num2str(generalized_density(W, g, team))]);        

            if sum( team'*M_lbs < lbs )
                display(['lower bound constraint is violated: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
            else
                display(['lower bound constraint is satisfied: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
            end
            if sum( team'*M_ubs > ubs )
                display(['upper bound constraint is violated: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
            else
                display(['upper bound constraint is satisfied: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
            end
            if sum(sum( max(0, dist(team>0, team>0) - d_0) )) ~= 0  %max(max(dist(team>0, team>0))) ~= 0
                display(['(Shifted/Scaled) Distance constraint is violated: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
            else
                display(['(Shifted/Scaled) Distance constraint is satisfied: Maximum distance in the team:', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
            end

            if debug
                disp(['****gamma= ', num2str(gamma)]);
                display(objectives);
                display(viols);
            end

        end


        %% Now choose the minimum objective among the nRuns for the
        %% current value of gamma.
        [min_obj, min_ix] = min(objectives(:, length(gammas)));
        cur_viols = viols(min_ix, length(gammas));
        satisfied = cur_viols == 0;
        last_cc_score = cc_score;

        disp(['******************** current number of violations = ', num2str(cur_viols), ', current gamma = ', num2str(gamma), ' ********************']);
        if debug
            disp(['**** min_ix= ', num2str(min_ix), '       min_obj= ', num2str(min_obj), '     cur_viols= ', num2str(cur_viols)]);
        end

        if satisfied
            % save the current best result!
            team = solutions(:, min_ix, length(gammas));
            cc_score = 1/objectives(min_ix, length(gammas));
            ip_opt = ip_opt_solutions(min_ix, length(gammas));

%             display(['All lower bound constraints are satisfied: Minimum skills of the team: ', num2str((team'*M_lbs).*max_M_lbs), ' Required: ', num2str(lbs.*max_M_lbs)]);
%             display(['All upper bound constraints are satisfied: Maximum skills/size/cost of the team: ', num2str((team'*M_ubs).*max_M_ubs), ' Required: ', num2str(ubs.*max_M_ubs)]);
%             display(['(Shifted/Scaled) Distance constraint is satisfied: Maximum distance in the team: ', num2str(max(max(dist(team>0, team>0)))), ' Required: ', num2str(d_0)]);
            disp('************* Finished solving an instance **************');
        else

            % initialize the new starts for the next gamma iteration
            % only if gamma >0 but keep the optimal solution for one
            % run
            %                 if gamma~=0
            %                     starts = solutions(:, :, length(gammas));
            %                 end

            if ~initialized
                starts(:, nRuns) = solutions(:, nRuns, length(gammas));
            end
            %                 if nRuns > 1
            %                     starts(:, nRuns) = solutions(:, nRuns, length(gammas));
            %                 end

            if prev_viols <= cur_viols %&& gamma <=10
                gamma_counter = gamma_counter*1.5;
            end
            prev_viols = cur_viols;

            gammas = [gammas gammas(end)+ gamma_counter];

        end

    end
    time_forte = toc(time_forte_start);

    % few assertions on the final solution...    
    assert( abs(generalized_density(W, g, team) - cc_score ) <= 1e-8 );
    assert( generalized_density(W, g, team) == objective_forte(W, g, team) );
    assert( sum( team'*M_lbs < lbs ) == 0 );
    assert( sum( team'*M_ubs > ubs ) == 0 );
    %assert( max(max(dist(team>0, team>0))) == 0 );
    assert( sum(sum( max(0, dist(team>0, team>0) - d_0) )) == 0);
    assert( sum(team(S)) == length(S) );

end
